@@ -1,7 +1,8 @@
+from __future__ import annotations
+
 import getpass
 import hashlib
 import json
-import mimetypes
 import os
 import pkgutil
 import re
@@ -9,20 +10,24 @@ import sys
 import time
 import typing as t
 import uuid
+from contextlib import ExitStack
+from io import BytesIO
 from itertools import chain
 from os.path import basename
 from os.path import join
+from zlib import adler32
 
 from .._internal import _log
+from ..exceptions import NotFound
 from ..http import parse_cookie
 from ..security import gen_salt
+from ..utils import send_file
 from ..wrappers.request import Request
 from ..wrappers.response import Response
 from .console import Console
-from .tbtools import Frame
-from .tbtools import get_current_traceback
+from .tbtools import DebugFrameSummary
+from .tbtools import DebugTraceback
 from .tbtools import render_console_html
-from .tbtools import Traceback
 
 if t.TYPE_CHECKING:
     from _typeshed.wsgi import StartResponse
@@ -37,16 +42,16 @@ def hash_pin(pin: str) -> str:
     return hashlib.sha1(f"{pin} added salt".encode("utf-8", "replace")).hexdigest()[:12]
 
 
-_machine_id: t.Optional[t.Union[str, bytes]] = None
+_machine_id: str | bytes | None = None
 
 
-def get_machine_id() -> t.Optional[t.Union[str, bytes]]:
+def get_machine_id() -> str | bytes | None:
     global _machine_id
 
     if _machine_id is not None:
         return _machine_id
 
-    def _generate() -> t.Optional[t.Union[str, bytes]]:
+    def _generate() -> str | bytes | None:
         linux = b""
 
         # machine-id is stable across boots, boot_id is not.
@@ -90,11 +95,9 @@ def get_machine_id() -> t.Optional[t.Union[str, bytes]]:
             pass
 
         # On Windows, use winreg to get the machine guid.
-        try:
+        if sys.platform == "win32":
             import winreg
-        except ImportError:
-            pass
-        else:
+
             try:
                 with winreg.OpenKey(
                     winreg.HKEY_LOCAL_MACHINE,
@@ -102,12 +105,12 @@ def get_machine_id() -> t.Optional[t.Union[str, bytes]]:
                     0,
                     winreg.KEY_READ | winreg.KEY_WOW64_64KEY,
                 ) as rk:
-                    guid: t.Union[str, bytes]
+                    guid: str | bytes
                     guid_type: int
                     guid, guid_type = winreg.QueryValueEx(rk, "MachineGuid")
 
                     if guid_type == winreg.REG_SZ:
-                        return guid.encode("utf-8")  # type: ignore
+                        return guid.encode("utf-8")
 
                     return guid
             except OSError:
@@ -124,14 +127,17 @@ class _ConsoleFrame:
     standalone console.
     """
 
-    def __init__(self, namespace: t.Dict[str, t.Any]):
+    def __init__(self, namespace: dict[str, t.Any]):
         self.console = Console(namespace)
         self.id = 0
 
+    def eval(self, code: str) -> t.Any:
+        return self.console.eval(code)
+
 
 def get_pin_and_cookie_name(
-    app: "WSGIApplication",
-) -> t.Union[t.Tuple[str, str], t.Tuple[None, None]]:
+    app: WSGIApplication,
+) -> tuple[str, str] | tuple[None, None]:
     """Given an application object this returns a semi-stable 9 digit pin
     code and a random key.  The hope is that this is stable between
     restarts to not make debugging particularly frustrating.  If the pin
@@ -148,7 +154,7 @@ def get_pin_and_cookie_name(
         return None, None
 
     # Pin was provided explicitly
-    if pin is not None and pin.replace("-", "").isdigit():
+    if pin is not None and pin.replace("-", "").isdecimal():
         # If there are separators in the pin, return it directly
         if "-" in pin:
             rv = pin
@@ -156,7 +162,7 @@ def get_pin_and_cookie_name(
             num = pin
 
     modname = getattr(app, "__module__", t.cast(object, app).__class__.__module__)
-    username: t.Optional[str]
+    username: str | None
 
     try:
         # getuser imports the pwd module, which does not exist in Google
@@ -222,13 +228,20 @@ class DebuggedApplication:
         from myapp import app
         app = DebuggedApplication(app, evalex=True)
 
-    The `evalex` keyword argument allows evaluating expressions in a
-    traceback's frame context.
+    The ``evalex`` argument allows evaluating expressions in any frame
+    of a traceback. This works by preserving each frame with its local
+    state. Some state, such as :doc:`local`, cannot be restored with the
+    frame by default. When ``evalex`` is enabled,
+    ``environ["werkzeug.debug.preserve_context"]`` will be a callable
+    that takes a context manager, and can be called multiple times.
+    Each context manager will be entered before evaluating code in the
+    frame, then exited again, so they can perform setup and cleanup for
+    each call.
 
     :param app: the WSGI application to run debugged.
     :param evalex: enable exception evaluation feature (interactive
                    debugging).  This requires a non-forking server.
-    :param request_key: The key that points to the request object in ths
+    :param request_key: The key that points to the request object in this
                         environment.  This parameter is ignored in current
                         versions.
     :param console_path: the URL for a general purpose console.
@@ -240,6 +253,9 @@ class DebuggedApplication:
                                to `True`.
     :param pin_security: can be used to disable the pin based security system.
     :param pin_logging: enables the logging of the pin system.
+
+    .. versionchanged:: 2.2
+        Added the ``werkzeug.debug.preserve_context`` environ key.
     """
 
     _pin: str
@@ -247,11 +263,11 @@ class DebuggedApplication:
 
     def __init__(
         self,
-        app: "WSGIApplication",
+        app: WSGIApplication,
         evalex: bool = False,
         request_key: str = "werkzeug.request",
         console_path: str = "/console",
-        console_init_func: t.Optional[t.Callable[[], t.Dict[str, t.Any]]] = None,
+        console_init_func: t.Callable[[], dict[str, t.Any]] | None = None,
         show_hidden_frames: bool = False,
         pin_security: bool = True,
         pin_logging: bool = True,
@@ -260,8 +276,8 @@ class DebuggedApplication:
             console_init_func = None
         self.app = app
         self.evalex = evalex
-        self.frames: t.Dict[int, t.Union[Frame, _ConsoleFrame]] = {}
-        self.tracebacks: t.Dict[int, Traceback] = {}
+        self.frames: dict[int, DebugFrameSummary | _ConsoleFrame] = {}
+        self.frame_contexts: dict[int, list[t.ContextManager[None]]] = {}
         self.request_key = request_key
         self.console_path = console_path
         self.console_init_func = console_init_func
@@ -282,7 +298,7 @@ class DebuggedApplication:
             self.pin = None
 
     @property
-    def pin(self) -> t.Optional[str]:
+    def pin(self) -> str | None:
         if not hasattr(self, "_pin"):
             pin_cookie = get_pin_and_cookie_name(self.app)
             self._pin, self._pin_cookie = pin_cookie  # type: ignore
@@ -301,37 +317,40 @@ class DebuggedApplication:
         return self._pin_cookie
 
     def debug_application(
-        self, environ: "WSGIEnvironment", start_response: "StartResponse"
+        self, environ: WSGIEnvironment, start_response: StartResponse
     ) -> t.Iterator[bytes]:
         """Run the application and conserve the traceback frames."""
+        contexts: list[t.ContextManager[t.Any]] = []
+
+        if self.evalex:
+            environ["werkzeug.debug.preserve_context"] = contexts.append
+
         app_iter = None
         try:
             app_iter = self.app(environ, start_response)
             yield from app_iter
             if hasattr(app_iter, "close"):
-                app_iter.close()  # type: ignore
-        except Exception:
+                app_iter.close()
+        except Exception as e:
             if hasattr(app_iter, "close"):
                 app_iter.close()  # type: ignore
-            traceback = get_current_traceback(
-                skip=1,
-                show_hidden_frames=self.show_hidden_frames,
-                ignore_system_exceptions=True,
+
+            tb = DebugTraceback(e, skip=1, hide=not self.show_hidden_frames)
+
+            for frame in tb.all_frames:
+                self.frames[id(frame)] = frame
+                self.frame_contexts[id(frame)] = contexts
+
+            is_trusted = bool(self.check_pin_trust(environ))
+            html = tb.render_debugger_html(
+                evalex=self.evalex,
+                secret=self.secret,
+                evalex_trusted=is_trusted,
             )
-            for frame in traceback.frames:
-                self.frames[frame.id] = frame
-            self.tracebacks[traceback.id] = traceback
+            response = Response(html, status=500, mimetype="text/html")
 
             try:
-                start_response(
-                    "500 INTERNAL SERVER ERROR",
-                    [
-                        ("Content-Type", "text/html; charset=utf-8"),
-                        # Disable Chrome's XSS protection, the debug
-                        # output can cause false-positives.
-                        ("X-XSS-Protection", "0"),
-                    ],
-                )
+                yield from response(environ, start_response)
             except Exception:
                 # if we end up here there has been output but an error
                 # occurred.  in that situation we can do nothing fancy any
@@ -342,19 +361,23 @@ class DebuggedApplication:
                     "response at a point where response headers were already "
                     "sent.\n"
                 )
-            else:
-                is_trusted = bool(self.check_pin_trust(environ))
-                yield traceback.render_full(
-                    evalex=self.evalex, evalex_trusted=is_trusted, secret=self.secret
-                ).encode("utf-8", "replace")
 
-            traceback.log(environ["wsgi.errors"])
+            environ["wsgi.errors"].write("".join(tb.render_traceback_text()))
 
-    def execute_command(
-        self, request: Request, command: str, frame: t.Union[Frame, _ConsoleFrame]
+    def execute_command(  # type: ignore[return]
+        self,
+        request: Request,
+        command: str,
+        frame: DebugFrameSummary | _ConsoleFrame,
     ) -> Response:
         """Execute a command in a console."""
-        return Response(frame.console.eval(command), mimetype="text/html")
+        contexts = self.frame_contexts.get(id(frame), [])
+
+        with ExitStack() as exit_stack:
+            for cm in contexts:
+                exit_stack.enter_context(cm)
+
+            return Response(frame.eval(command), mimetype="text/html")
 
     def display_console(self, request: Request) -> Response:
         """Display a standalone shell."""
@@ -373,17 +396,22 @@ class DebuggedApplication:
 
     def get_resource(self, request: Request, filename: str) -> Response:
         """Return a static resource from the shared folder."""
-        filename = join("shared", basename(filename))
-        try:
-            data = pkgutil.get_data(__package__, filename)
-        except OSError:
-            data = None
-        if data is not None:
-            mimetype = mimetypes.guess_type(filename)[0] or "application/octet-stream"
-            return Response(data, mimetype=mimetype)
-        return Response("Not Found", status=404)
+        path = join("shared", basename(filename))
 
-    def check_pin_trust(self, environ: "WSGIEnvironment") -> t.Optional[bool]:
+        try:
+            data = pkgutil.get_data(__package__, path)
+        except OSError:
+            return NotFound()  # type: ignore[return-value]
+        else:
+            if data is None:
+                return NotFound()  # type: ignore[return-value]
+
+            etag = str(adler32(data) & 0xFFFFFFFF)
+            return send_file(
+                BytesIO(data), request.environ, download_name=filename, etag=etag
+            )
+
+    def check_pin_trust(self, environ: WSGIEnvironment) -> bool | None:
         """Checks if the request passed the pin test.  This returns `True` if the
         request is trusted on a pin/cookie basis and returns `False` if not.
         Additionally if the cookie's stored pin hash is wrong it will return
@@ -394,12 +422,16 @@ class DebuggedApplication:
         val = parse_cookie(environ).get(self.pin_cookie_name)
         if not val or "|" not in val:
             return False
-        ts, pin_hash = val.split("|", 1)
-        if not ts.isdigit():
+        ts_str, pin_hash = val.split("|", 1)
+
+        try:
+            ts = int(ts_str)
+        except ValueError:
             return False
+
         if pin_hash != hash_pin(self.pin):
             return None
-        return (time.time() - PIN_TIME) < int(ts)
+        return (time.time() - PIN_TIME) < ts
 
     def _fail_pin_auth(self) -> None:
         time.sleep(5.0 if self._failed_pin_auth > 5 else 0.5)
@@ -449,7 +481,8 @@ class DebuggedApplication:
                 self.pin_cookie_name,
                 f"{int(time.time())}|{hash_pin(pin)}",
                 httponly=True,
-                samesite="None",
+                samesite="Strict",
+                secure=request.is_secure,
             )
         elif bad_cookie:
             rv.delete_cookie(self.pin_cookie_name)
@@ -465,7 +498,7 @@ class DebuggedApplication:
         return Response("")
 
     def __call__(
-        self, environ: "WSGIEnvironment", start_response: "StartResponse"
+        self, environ: WSGIEnvironment, start_response: StartResponse
     ) -> t.Iterable[bytes]:
         """Dispatch the requests."""
         # important: don't ever access a function here that reads the incoming
