@@ -1,15 +1,14 @@
+from __future__ import annotations
+
 import typing as t
 import warnings
-from functools import update_wrapper
 from io import BytesIO
-from itertools import chain
-from typing import Union
+from urllib.parse import parse_qsl
 
-from . import exceptions
-from ._internal import _to_str
 from .datastructures import FileStorage
 from .datastructures import Headers
 from .datastructures import MultiDict
+from .exceptions import RequestEntityTooLarge
 from .http import parse_options_header
 from .sansio.multipart import Data
 from .sansio.multipart import Epilogue
@@ -17,8 +16,6 @@ from .sansio.multipart import Field
 from .sansio.multipart import File
 from .sansio.multipart import MultipartDecoder
 from .sansio.multipart import NeedData
-from .urls import url_decode_stream
-from .wsgi import _make_chunk_iter
 from .wsgi import get_content_length
 from .wsgi import get_input_stream
 
@@ -32,56 +29,53 @@ except ImportError:
     SpooledTemporaryFile = None  # type: ignore
 
 if t.TYPE_CHECKING:
-    from wsgiref.types import WSGIEnvironment
+    import typing as te
+    from _typeshed.wsgi import WSGIEnvironment
 
-    t_parse_result = t.Tuple[t.BinaryIO, MultiDict, MultiDict]
+    t_parse_result = t.Tuple[t.IO[bytes], MultiDict, MultiDict]
 
-    class TStreamFactory(t.Protocol):
+    class TStreamFactory(te.Protocol):
         def __call__(
             self,
-            total_content_length: t.Optional[int],
-            content_type: t.Optional[str],
-            filename: t.Optional[str],
-            content_length: t.Optional[int] = None,
-        ) -> t.BinaryIO:
+            total_content_length: int | None,
+            content_type: str | None,
+            filename: str | None,
+            content_length: int | None = None,
+        ) -> t.IO[bytes]:
             ...
 
 
 F = t.TypeVar("F", bound=t.Callable[..., t.Any])
 
 
-def _exhaust(stream: t.BinaryIO) -> None:
-    bts = stream.read(64 * 1024)
-    while bts:
-        bts = stream.read(64 * 1024)
-
-
 def default_stream_factory(
-    total_content_length: t.Optional[int],
-    content_type: t.Optional[str],
-    filename: t.Optional[str],
-    content_length: t.Optional[int] = None,
-) -> t.BinaryIO:
+    total_content_length: int | None,
+    content_type: str | None,
+    filename: str | None,
+    content_length: int | None = None,
+) -> t.IO[bytes]:
     max_size = 1024 * 500
 
     if SpooledTemporaryFile is not None:
-        return t.cast(t.BinaryIO, SpooledTemporaryFile(max_size=max_size, mode="rb+"))
+        return t.cast(t.IO[bytes], SpooledTemporaryFile(max_size=max_size, mode="rb+"))
     elif total_content_length is None or total_content_length > max_size:
-        return t.cast(t.BinaryIO, TemporaryFile("rb+"))
+        return t.cast(t.IO[bytes], TemporaryFile("rb+"))
 
     return BytesIO()
 
 
 def parse_form_data(
-    environ: "WSGIEnvironment",
-    stream_factory: t.Optional["TStreamFactory"] = None,
-    charset: str = "utf-8",
-    errors: str = "replace",
-    max_form_memory_size: t.Optional[int] = None,
-    max_content_length: t.Optional[int] = None,
-    cls: t.Optional[t.Type[MultiDict]] = None,
+    environ: WSGIEnvironment,
+    stream_factory: TStreamFactory | None = None,
+    charset: str | None = None,
+    errors: str | None = None,
+    max_form_memory_size: int | None = None,
+    max_content_length: int | None = None,
+    cls: type[MultiDict] | None = None,
     silent: bool = True,
-) -> "t_parse_result":
+    *,
+    max_form_parts: int | None = None,
+) -> t_parse_result:
     """Parse the form data in the environ and return it as tuple in the form
     ``(stream, form, files)``.  You should only call this method if the
     transport method is `POST`, `PUT`, or `PATCH`.
@@ -95,19 +89,10 @@ def parse_form_data(
 
     Have a look at :doc:`/request_data` for more details.
 
-    .. versionadded:: 0.5
-       The `max_form_memory_size`, `max_content_length` and
-       `cls` parameters were added.
-
-    .. versionadded:: 0.5.1
-       The optional `silent` flag was added.
-
     :param environ: the WSGI environment to be used for parsing.
     :param stream_factory: An optional callable that returns a new read and
                            writeable file descriptor.  This callable works
                            the same as :meth:`Response._get_file_stream`.
-    :param charset: The character set for URL and url encoded form data.
-    :param errors: The encoding error behavior.
     :param max_form_memory_size: the maximum number of bytes to be accepted for
                            in-memory stored form data.  If the data
                            exceeds the value specified an
@@ -120,40 +105,36 @@ def parse_form_data(
     :param cls: an optional dict class to use.  If this is not specified
                        or `None` the default :class:`MultiDict` is used.
     :param silent: If set to False parsing errors will not be caught.
+    :param max_form_parts: The maximum number of parts to be parsed. If this is
+        exceeded, a :exc:`~exceptions.RequestEntityTooLarge` exception is raised.
     :return: A tuple in the form ``(stream, form, files)``.
+
+    .. versionchanged:: 2.3
+        Added the ``max_form_parts`` parameter.
+
+    .. versionchanged:: 2.3
+        The ``charset`` and ``errors`` parameters are deprecated and will be removed in
+        Werkzeug 2.4.
+
+    .. versionadded:: 0.5.1
+       Added the ``silent`` parameter.
+
+    .. versionadded:: 0.5
+       Added the ``max_form_memory_size``, ``max_content_length``, and ``cls``
+       parameters.
     """
     return FormDataParser(
-        stream_factory,
-        charset,
-        errors,
-        max_form_memory_size,
-        max_content_length,
-        cls,
-        silent,
+        stream_factory=stream_factory,
+        charset=charset,
+        errors=errors,
+        max_form_memory_size=max_form_memory_size,
+        max_content_length=max_content_length,
+        max_form_parts=max_form_parts,
+        silent=silent,
+        cls=cls,
     ).parse_from_environ(environ)
 
 
-def exhaust_stream(f: F) -> F:
-    """Helper decorator for methods that exhausts the stream on return."""
-
-    def wrapper(self, stream, *args, **kwargs):  # type: ignore
-        try:
-            return f(self, stream, *args, **kwargs)
-        finally:
-            exhaust = getattr(stream, "exhaust", None)
-
-            if exhaust is not None:
-                exhaust()
-            else:
-                while True:
-                    chunk = stream.read(1024 * 64)
-
-                    if not chunk:
-                        break
-
-    return update_wrapper(t.cast(F, wrapper), f)
-
-
 class FormDataParser:
     """This class implements parsing of form data for Werkzeug.  By itself
     it can parse multipart and url encoded form data.  It can be subclassed
@@ -161,13 +142,9 @@ class FormDataParser:
     untouched stream and expose it as separate attributes on a request
     object.
 
-    .. versionadded:: 0.8
-
     :param stream_factory: An optional callable that returns a new read and
                            writeable file descriptor.  This callable works
                            the same as :meth:`Response._get_file_stream`.
-    :param charset: The character set for URL and url encoded form data.
-    :param errors: The encoding error behavior.
     :param max_form_memory_size: the maximum number of bytes to be accepted for
                            in-memory stored form data.  If the data
                            exceeds the value specified an
@@ -180,26 +157,66 @@ class FormDataParser:
     :param cls: an optional dict class to use.  If this is not specified
                        or `None` the default :class:`MultiDict` is used.
     :param silent: If set to False parsing errors will not be caught.
+    :param max_form_parts: The maximum number of parts to be parsed. If this is
+        exceeded, a :exc:`~exceptions.RequestEntityTooLarge` exception is raised.
+
+    .. versionchanged:: 2.3
+        The ``charset`` and ``errors`` parameters are deprecated and will be removed in
+        Werkzeug 2.4.
+
+    .. versionchanged:: 2.3
+        The ``parse_functions`` attribute and ``get_parse_func`` methods are deprecated
+        and will be removed in Werkzeug 2.4.
+
+    .. versionchanged:: 2.2.3
+        Added the ``max_form_parts`` parameter.
+
+    .. versionadded:: 0.8
     """
 
     def __init__(
         self,
-        stream_factory: t.Optional["TStreamFactory"] = None,
-        charset: str = "utf-8",
-        errors: str = "replace",
-        max_form_memory_size: t.Optional[int] = None,
-        max_content_length: t.Optional[int] = None,
-        cls: t.Optional[t.Type[MultiDict]] = None,
+        stream_factory: TStreamFactory | None = None,
+        charset: str | None = None,
+        errors: str | None = None,
+        max_form_memory_size: int | None = None,
+        max_content_length: int | None = None,
+        cls: type[MultiDict] | None = None,
         silent: bool = True,
+        *,
+        max_form_parts: int | None = None,
     ) -> None:
         if stream_factory is None:
             stream_factory = default_stream_factory
 
         self.stream_factory = stream_factory
+
+        if charset is not None:
+            warnings.warn(
+                "The 'charset' parameter is deprecated and will be"
+                " removed in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        else:
+            charset = "utf-8"
+
         self.charset = charset
+
+        if errors is not None:
+            warnings.warn(
+                "The 'errors' parameter is deprecated and will be"
+                " removed in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        else:
+            errors = "replace"
+
         self.errors = errors
         self.max_form_memory_size = max_form_memory_size
         self.max_content_length = max_content_length
+        self.max_form_parts = max_form_parts
 
         if cls is None:
             cls = MultiDict
@@ -208,33 +225,66 @@ class FormDataParser:
         self.silent = silent
 
     def get_parse_func(
-        self, mimetype: str, options: t.Dict[str, str]
-    ) -> t.Optional[
+        self, mimetype: str, options: dict[str, str]
+    ) -> None | (
         t.Callable[
-            ["FormDataParser", t.BinaryIO, str, t.Optional[int], t.Dict[str, str]],
-            "t_parse_result",
+            [FormDataParser, t.IO[bytes], str, int | None, dict[str, str]],
+            t_parse_result,
         ]
-    ]:
-        return self.parse_functions.get(mimetype)
+    ):
+        warnings.warn(
+            "The 'get_parse_func' method is deprecated and will be"
+            " removed in Werkzeug 2.4.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
 
-    def parse_from_environ(self, environ: "WSGIEnvironment") -> "t_parse_result":
+        if mimetype == "multipart/form-data":
+            return type(self)._parse_multipart
+        elif mimetype == "application/x-www-form-urlencoded":
+            return type(self)._parse_urlencoded
+        elif mimetype == "application/x-url-encoded":
+            warnings.warn(
+                "The 'application/x-url-encoded' mimetype is invalid, and will not be"
+                " treated as 'application/x-www-form-urlencoded' in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+            return type(self)._parse_urlencoded
+        elif mimetype in self.parse_functions:
+            warnings.warn(
+                "The 'parse_functions' attribute is deprecated and will be removed in"
+                " Werkzeug 2.4. Override 'parse' instead.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+            return self.parse_functions[mimetype]
+
+        return None
+
+    def parse_from_environ(self, environ: WSGIEnvironment) -> t_parse_result:
         """Parses the information from the environment as form data.
 
         :param environ: the WSGI environment to be used for parsing.
         :return: A tuple in the form ``(stream, form, files)``.
         """
-        content_type = environ.get("CONTENT_TYPE", "")
+        stream = get_input_stream(environ, max_content_length=self.max_content_length)
         content_length = get_content_length(environ)
-        mimetype, options = parse_options_header(content_type)
-        return self.parse(get_input_stream(environ), mimetype, content_length, options)
+        mimetype, options = parse_options_header(environ.get("CONTENT_TYPE"))
+        return self.parse(
+            stream,
+            content_length=content_length,
+            mimetype=mimetype,
+            options=options,
+        )
 
     def parse(
         self,
-        stream: t.BinaryIO,
+        stream: t.IO[bytes],
         mimetype: str,
-        content_length: t.Optional[int],
-        options: t.Optional[t.Dict[str, str]] = None,
-    ) -> "t_parse_result":
+        content_length: int | None,
+        options: dict[str, str] | None = None,
+    ) -> t_parse_result:
         """Parses the information from the given stream, mimetype,
         content length and mimetype parameters.
 
@@ -244,43 +294,60 @@ class FormDataParser:
         :param options: optional mimetype parameters (used for
                         the multipart boundary for instance)
         :return: A tuple in the form ``(stream, form, files)``.
+
+        .. versionchanged:: 2.3
+            The ``application/x-url-encoded`` content type is deprecated and will not be
+            treated as ``application/x-www-form-urlencoded`` in Werkzeug 2.4.
         """
-        if (
-            self.max_content_length is not None
-            and content_length is not None
-            and content_length > self.max_content_length
-        ):
-            # if the input stream is not exhausted, firefox reports Connection Reset
-            _exhaust(stream)
-            raise exceptions.RequestEntityTooLarge()
+        if mimetype == "multipart/form-data":
+            parse_func = self._parse_multipart
+        elif mimetype == "application/x-www-form-urlencoded":
+            parse_func = self._parse_urlencoded
+        elif mimetype == "application/x-url-encoded":
+            warnings.warn(
+                "The 'application/x-url-encoded' mimetype is invalid, and will not be"
+                " treated as 'application/x-www-form-urlencoded' in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+            parse_func = self._parse_urlencoded
+        elif mimetype in self.parse_functions:
+            warnings.warn(
+                "The 'parse_functions' attribute is deprecated and will be removed in"
+                " Werkzeug 2.4. Override 'parse' instead.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+            parse_func = self.parse_functions[mimetype].__get__(self, type(self))
+        else:
+            return stream, self.cls(), self.cls()
 
         if options is None:
             options = {}
 
-        parse_func = self.get_parse_func(mimetype, options)
-
-        if parse_func is not None:
-            try:
-                return parse_func(self, stream, mimetype, content_length, options)
-            except ValueError:
-                if not self.silent:
-                    raise
+        try:
+            return parse_func(stream, mimetype, content_length, options)
+        except ValueError:
+            if not self.silent:
+                raise
 
         return stream, self.cls(), self.cls()
 
-    @exhaust_stream
     def _parse_multipart(
         self,
-        stream: t.BinaryIO,
+        stream: t.IO[bytes],
         mimetype: str,
-        content_length: t.Optional[int],
-        options: t.Dict[str, str],
-    ) -> "t_parse_result":
+        content_length: int | None,
+        options: dict[str, str],
+    ) -> t_parse_result:
+        charset = self.charset if self.charset != "utf-8" else None
+        errors = self.errors if self.errors != "replace" else None
         parser = MultiPartParser(
-            self.stream_factory,
-            self.charset,
-            self.errors,
+            stream_factory=self.stream_factory,
+            charset=charset,
+            errors=errors,
             max_form_memory_size=self.max_form_memory_size,
+            max_form_parts=self.max_form_parts,
             cls=self.cls,
         )
         boundary = options.get("boundary", "").encode("ascii")
@@ -291,104 +358,78 @@ class FormDataParser:
         form, files = parser.parse(stream, boundary, content_length)
         return stream, form, files
 
-    @exhaust_stream
     def _parse_urlencoded(
         self,
-        stream: t.BinaryIO,
+        stream: t.IO[bytes],
         mimetype: str,
-        content_length: t.Optional[int],
-        options: t.Dict[str, str],
-    ) -> "t_parse_result":
+        content_length: int | None,
+        options: dict[str, str],
+    ) -> t_parse_result:
         if (
             self.max_form_memory_size is not None
             and content_length is not None
             and content_length > self.max_form_memory_size
         ):
-            # if the input stream is not exhausted, firefox reports Connection Reset
-            _exhaust(stream)
-            raise exceptions.RequestEntityTooLarge()
+            raise RequestEntityTooLarge()
 
-        form = url_decode_stream(stream, self.charset, errors=self.errors, cls=self.cls)
-        return stream, form, self.cls()
+        try:
+            items = parse_qsl(
+                stream.read().decode(),
+                keep_blank_values=True,
+                encoding=self.charset,
+                errors="werkzeug.url_quote",
+                max_num_fields=self.max_form_parts,
+            )
+        except ValueError as e:
+            raise RequestEntityTooLarge() from e
 
-    #: mapping of mimetypes to parsing functions
-    parse_functions: t.Dict[
+        return stream, self.cls(items), self.cls()
+
+    parse_functions: dict[
         str,
         t.Callable[
-            ["FormDataParser", t.BinaryIO, str, t.Optional[int], t.Dict[str, str]],
-            "t_parse_result",
+            [FormDataParser, t.IO[bytes], str, int | None, dict[str, str]],
+            t_parse_result,
         ],
-    ] = {
-        "multipart/form-data": _parse_multipart,
-        "application/x-www-form-urlencoded": _parse_urlencoded,
-        "application/x-url-encoded": _parse_urlencoded,
-    }
-
-
-def _line_parse(line: str) -> t.Tuple[str, bool]:
-    """Removes line ending characters and returns a tuple (`stripped_line`,
-    `is_terminated`).
-    """
-    if line[-2:] == "\r\n":
-        return line[:-2], True
-
-    elif line[-1:] in {"\r", "\n"}:
-        return line[:-1], True
-
-    return line, False
-
-
-def parse_multipart_headers(iterable: t.Iterable[bytes]) -> Headers:
-    """Parses multipart headers from an iterable that yields lines (including
-    the trailing newline symbol).  The iterable has to be newline terminated.
-    The iterable will stop at the line where the headers ended so it can be
-    further consumed.
-    :param iterable: iterable of strings that are newline terminated
-    """
-    warnings.warn(
-        "'parse_multipart_headers' is deprecated and will be removed in"
-        " Werkzeug 2.1.",
-        DeprecationWarning,
-        stacklevel=2,
-    )
-    result: t.List[t.Tuple[str, str]] = []
-
-    for b_line in iterable:
-        line = _to_str(b_line)
-        line, line_terminated = _line_parse(line)
-
-        if not line_terminated:
-            raise ValueError("unexpected end of line in multipart header")
-
-        if not line:
-            break
-        elif line[0] in " \t" and result:
-            key, value = result[-1]
-            result[-1] = (key, f"{value}\n {line[1:]}")
-        else:
-            parts = line.split(":", 1)
-
-            if len(parts) == 2:
-                result.append((parts[0].strip(), parts[1].strip()))
-
-    # we link the list to the headers, no need to create a copy, the
-    # list was not shared anyways.
-    return Headers(result)
+    ] = {}
 
 
 class MultiPartParser:
     def __init__(
         self,
-        stream_factory: t.Optional["TStreamFactory"] = None,
-        charset: str = "utf-8",
-        errors: str = "replace",
-        max_form_memory_size: t.Optional[int] = None,
-        cls: t.Optional[t.Type[MultiDict]] = None,
+        stream_factory: TStreamFactory | None = None,
+        charset: str | None = None,
+        errors: str | None = None,
+        max_form_memory_size: int | None = None,
+        cls: type[MultiDict] | None = None,
         buffer_size: int = 64 * 1024,
+        max_form_parts: int | None = None,
     ) -> None:
+        if charset is not None:
+            warnings.warn(
+                "The 'charset' parameter is deprecated and will be"
+                " removed in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        else:
+            charset = "utf-8"
+
         self.charset = charset
+
+        if errors is not None:
+            warnings.warn(
+                "The 'errors' parameter is deprecated and will be"
+                " removed in Werkzeug 2.4.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+        else:
+            errors = "replace"
+
         self.errors = errors
         self.max_form_memory_size = max_form_memory_size
+        self.max_form_parts = max_form_parts
 
         if stream_factory is None:
             stream_factory = default_stream_factory
@@ -399,10 +440,9 @@ class MultiPartParser:
             cls = MultiDict
 
         self.cls = cls
-
         self.buffer_size = buffer_size
 
-    def fail(self, message: str) -> "t.NoReturn":
+    def fail(self, message: str) -> te.NoReturn:
         raise ValueError(message)
 
     def get_part_charset(self, headers: Headers) -> str:
@@ -410,14 +450,19 @@ class MultiPartParser:
         content_type = headers.get("content-type")
 
         if content_type:
-            mimetype, ct_params = parse_options_header(content_type)
-            return ct_params.get("charset", self.charset)
+            parameters = parse_options_header(content_type)[1]
+            ct_charset = parameters.get("charset", "").lower()
+
+            # A safe list of encodings. Modern clients should only send ASCII or UTF-8.
+            # This list will not be extended further.
+            if ct_charset in {"ascii", "us-ascii", "utf-8", "iso-8859-1"}:
+                return ct_charset
 
         return self.charset
 
     def start_file_streaming(
-        self, event: File, total_content_length: t.Optional[int]
-    ) -> t.BinaryIO:
+        self, event: File, total_content_length: int | None
+    ) -> t.IO[bytes]:
         content_type = event.headers.get("content-type")
 
         try:
@@ -434,27 +479,22 @@ class MultiPartParser:
         return container
 
     def parse(
-        self, stream: t.BinaryIO, boundary: bytes, content_length: t.Optional[int]
-    ) -> t.Tuple[MultiDict, MultiDict]:
-        container: t.Union[t.BinaryIO, t.List[bytes]]
+        self, stream: t.IO[bytes], boundary: bytes, content_length: int | None
+    ) -> tuple[MultiDict, MultiDict]:
+        current_part: Field | File
+        container: t.IO[bytes] | list[bytes]
         _write: t.Callable[[bytes], t.Any]
 
-        iterator = chain(
-            _make_chunk_iter(
-                stream,
-                limit=content_length,
-                buffer_size=self.buffer_size,
-            ),
-            [None],
+        parser = MultipartDecoder(
+            boundary,
+            max_form_memory_size=self.max_form_memory_size,
+            max_parts=self.max_form_parts,
         )
 
-        parser = MultipartDecoder(boundary, self.max_form_memory_size)
-
         fields = []
         files = []
 
-        current_part: Union[Field, File]
-        for data in iterator:
+        for data in _chunk_iter(stream.read, self.buffer_size):
             parser.receive_data(data)
             event = parser.next_event()
             while not isinstance(event, (Epilogue, NeedData)):
@@ -475,7 +515,7 @@ class MultiPartParser:
                             )
                             fields.append((current_part.name, value))
                         else:
-                            container = t.cast(t.BinaryIO, container)
+                            container = t.cast(t.IO[bytes], container)
                             container.seek(0)
                             files.append(
                                 (
@@ -492,3 +532,18 @@ class MultiPartParser:
                 event = parser.next_event()
 
         return self.cls(fields), self.cls(files)
+
+
+def _chunk_iter(read: t.Callable[[int], bytes], size: int) -> t.Iterator[bytes | None]:
+    """Read data in chunks for multipart/form-data parsing. Stop if no data is read.
+    Yield ``None`` at the end to signal end of parsing.
+    """
+    while True:
+        data = read(size)
+
+        if not data:
+            break
+
+        yield data
+
+    yield None
