@@ -1,14 +1,11 @@
+from __future__ import annotations
+
 import re
+import typing as t
 from dataclasses import dataclass
 from enum import auto
 from enum import Enum
-from typing import cast
-from typing import List
-from typing import Optional
-from typing import Tuple
 
-from .._internal import _to_bytes
-from .._internal import _to_str
 from ..datastructures import Headers
 from ..exceptions import RequestEntityTooLarge
 from ..http import parse_options_header
@@ -58,6 +55,7 @@ class State(Enum):
     PREAMBLE = auto()
     PART = auto()
     DATA = auto()
+    DATA_START = auto()
     EPILOGUE = auto()
     COMPLETE = auto()
 
@@ -70,6 +68,10 @@ LINE_BREAK_RE = re.compile(LINE_BREAK, re.MULTILINE)
 # Header values can be continued via a space or tab after the linebreak, as
 # per RFC2231
 HEADER_CONTINUATION_RE = re.compile(b"%s[ \t]" % LINE_BREAK, re.MULTILINE)
+# This must be long enough to contain any line breaks plus any
+# additional boundary markers (--) such that they will be found in a
+# subsequent search
+SEARCH_EXTRA_LENGTH = 8
 
 
 class MultipartDecoder:
@@ -82,11 +84,14 @@ class MultipartDecoder:
     def __init__(
         self,
         boundary: bytes,
-        max_form_memory_size: Optional[int] = None,
+        max_form_memory_size: int | None = None,
+        *,
+        max_parts: int | None = None,
     ) -> None:
         self.buffer = bytearray()
         self.complete = False
         self.max_form_memory_size = max_form_memory_size
+        self.max_parts = max_parts
         self.state = State.PREAMBLE
         self.boundary = boundary
 
@@ -100,7 +105,7 @@ class MultipartDecoder:
         # epilogue boundary (for empty form-data) hence the matching
         # group to understand if it is an epilogue boundary.
         self.preamble_re = re.compile(
-            br"%s?--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
+            rb"%s?--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
             % (LINE_BREAK, re.escape(boundary), LINE_BREAK, LINE_BREAK),
             re.MULTILINE,
         )
@@ -109,10 +114,12 @@ class MultipartDecoder:
         # could be the epilogue boundary hence the matching group to
         # understand if it is an epilogue boundary.
         self.boundary_re = re.compile(
-            br"%s--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
+            rb"%s--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
             % (LINE_BREAK, re.escape(boundary), LINE_BREAK, LINE_BREAK),
             re.MULTILINE,
         )
+        self._search_position = 0
+        self._parts_decoded = 0
 
     def last_newline(self) -> int:
         try:
@@ -126,7 +133,7 @@ class MultipartDecoder:
 
         return min(last_nl, last_cr)
 
-    def receive_data(self, data: Optional[bytes]) -> None:
+    def receive_data(self, data: bytes | None) -> None:
         if data is None:
             self.complete = True
         elif (
@@ -141,7 +148,7 @@ class MultipartDecoder:
         event: Event = NEED_DATA
 
         if self.state == State.PREAMBLE:
-            match = self.preamble_re.search(self.buffer)
+            match = self.preamble_re.search(self.buffer, self._search_position)
             if match is not None:
                 if match.group(1).startswith(b"--"):
                     self.state = State.EPILOGUE
@@ -150,12 +157,24 @@ class MultipartDecoder:
                 data = bytes(self.buffer[: match.start()])
                 del self.buffer[: match.end()]
                 event = Preamble(data=data)
+                self._search_position = 0
+            else:
+                # Update the search start position to be equal to the
+                # current buffer length (already searched) minus a
+                # safe buffer for part of the search target.
+                self._search_position = max(
+                    0, len(self.buffer) - len(self.boundary) - SEARCH_EXTRA_LENGTH
+                )
 
         elif self.state == State.PART:
-            match = BLANK_LINE_RE.search(self.buffer)
+            match = BLANK_LINE_RE.search(self.buffer, self._search_position)
             if match is not None:
                 headers = self._parse_headers(self.buffer[: match.start()])
-                del self.buffer[: match.end()]
+                # The final header ends with a single CRLF, however a
+                # blank line indicates the start of the
+                # body. Therefore the end is after the first CRLF.
+                headers_end = (match.start() + match.end()) // 2
+                del self.buffer[:headers_end]
 
                 if "content-disposition" not in headers:
                     raise ValueError("Missing Content-Disposition header")
@@ -163,7 +182,7 @@ class MultipartDecoder:
                 disposition, extra = parse_options_header(
                     headers["content-disposition"]
                 )
-                name = cast(str, extra.get("name"))
+                name = t.cast(str, extra.get("name"))
                 filename = extra.get("filename")
                 if filename is not None:
                     event = File(
@@ -176,30 +195,27 @@ class MultipartDecoder:
                         headers=headers,
                         name=name,
                     )
+                self.state = State.DATA_START
+                self._search_position = 0
+                self._parts_decoded += 1
+
+                if self.max_parts is not None and self._parts_decoded > self.max_parts:
+                    raise RequestEntityTooLarge()
+            else:
+                # Update the search start position to be equal to the
+                # current buffer length (already searched) minus a
+                # safe buffer for part of the search target.
+                self._search_position = max(0, len(self.buffer) - SEARCH_EXTRA_LENGTH)
+
+        elif self.state == State.DATA_START:
+            data, del_index, more_data = self._parse_data(self.buffer, start=True)
+            del self.buffer[:del_index]
+            event = Data(data=data, more_data=more_data)
+            if more_data:
                 self.state = State.DATA
 
         elif self.state == State.DATA:
-            if self.buffer.find(b"--" + self.boundary) == -1:
-                # No complete boundary in the buffer, but there may be
-                # a partial boundary at the end. As the boundary
-                # starts with either a nl or cr find the earliest and
-                # return up to that as data.
-                data_length = del_index = self.last_newline()
-                more_data = True
-            else:
-                match = self.boundary_re.search(self.buffer)
-                if match is not None:
-                    if match.group(1).startswith(b"--"):
-                        self.state = State.EPILOGUE
-                    else:
-                        self.state = State.PART
-                    data_length = match.start()
-                    del_index = match.end()
-                else:
-                    data_length = del_index = self.last_newline()
-                more_data = match is None
-
-            data = bytes(self.buffer[:data_length])
+            data, del_index, more_data = self._parse_data(self.buffer, start=False)
             del self.buffer[:del_index]
             if data or not more_data:
                 event = Data(data=data, more_data=more_data)
@@ -215,16 +231,40 @@ class MultipartDecoder:
         return event
 
     def _parse_headers(self, data: bytes) -> Headers:
-        headers: List[Tuple[str, str]] = []
+        headers: list[tuple[str, str]] = []
         # Merge the continued headers into one line
         data = HEADER_CONTINUATION_RE.sub(b" ", data)
         # Now there is one header per line
         for line in data.splitlines():
-            if line.strip() != b"":
-                name, value = _to_str(line).strip().split(":", 1)
+            line = line.strip()
+
+            if line != b"":
+                name, _, value = line.decode().partition(":")
                 headers.append((name.strip(), value.strip()))
         return Headers(headers)
 
+    def _parse_data(self, data: bytes, *, start: bool) -> tuple[bytes, int, bool]:
+        # Body parts must start with CRLF (or CR or LF)
+        if start:
+            match = LINE_BREAK_RE.match(data)
+            data_start = t.cast(t.Match[bytes], match).end()
+        else:
+            data_start = 0
+
+        match = self.boundary_re.search(data)
+        if match is not None:
+            if match.group(1).startswith(b"--"):
+                self.state = State.EPILOGUE
+            else:
+                self.state = State.PART
+            data_end = match.start()
+            del_index = match.end()
+        else:
+            data_end = del_index = self.last_newline()
+        more_data = match is None
+
+        return bytes(data[data_start:data_end]), del_index, more_data
+
 
 class MultipartEncoder:
     def __init__(self, boundary: bytes) -> None:
@@ -240,17 +280,22 @@ class MultipartEncoder:
             State.PART,
             State.DATA,
         }:
-            self.state = State.DATA
             data = b"\r\n--" + self.boundary + b"\r\n"
-            data += b'Content-Disposition: form-data; name="%s"' % _to_bytes(event.name)
+            data += b'Content-Disposition: form-data; name="%s"' % event.name.encode()
             if isinstance(event, File):
-                data += b'; filename="%s"' % _to_bytes(event.filename)
+                data += b'; filename="%s"' % event.filename.encode()
             data += b"\r\n"
-            for name, value in cast(Field, event).headers:
+            for name, value in t.cast(Field, event).headers:
                 if name.lower() != "content-disposition":
-                    data += _to_bytes(f"{name}: {value}\r\n")
-            data += b"\r\n"
+                    data += f"{name}: {value}\r\n".encode()
+            self.state = State.DATA_START
             return data
+        elif isinstance(event, Data) and self.state == State.DATA_START:
+            self.state = State.DATA
+            if len(event.data) > 0:
+                return b"\r\n" + event.data
+            else:
+                return event.data
         elif isinstance(event, Data) and self.state == State.DATA:
             return event.data
         elif isinstance(event, Epilogue):
